class ClassNotFoundException(Exception):
    def __init__(self, cls: str, *args):
        super().__init__(f'Class "{cls}" is not found.', *args)


class ContextNotLoadedError(RuntimeError):
    pass



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import BaseConfig, ConfigListStructure


class ConfigStructureVisitor(ABC):
    @abstractmethod
    def visit_config(self, structure: "BaseConfig") -> None: ...

    @abstractmethod
    def visit_config_list(self, structure: "ConfigListStructure") -> None: ...



from argparse import Namespace
from pathlib import Path
from typing import Dict, TYPE_CHECKING, Union

import yaml
from abc import ABC, abstractmethod

from . import attribute_setters

if TYPE_CHECKING:
    from ..config import BaseConfig

setter_name_map = {}

for name, x in vars(attribute_setters).items():
    if isinstance(x, type) and issubclass(x, attribute_setters.AttributeSetter):
        try:
            name = x.get_setter_name()
            setter_name_map[name] = x()
        except NotImplementedError:
            pass


class BaseConfigLoader(ABC):
    _attribute_setter: attribute_setters.AttributeSetter

    def __init__(self, setter: Union[attribute_setters.AttributeSetter, str] = None):
        setter = "strict" if setter is None else setter
        if isinstance(setter, str):
            self._attribute_setter = setter_name_map[setter]
        else:
            self._attribute_setter = setter

    @abstractmethod
    def load(self, config: 'BaseConfig'):
        ...

    @abstractmethod
    def get_sub_loader(self, val) -> "BaseConfigLoader":
        ...

    def get_setter(self) -> attribute_setters.AttributeSetter:
        return self._attribute_setter


class DictBasedConfigLoader(BaseConfigLoader, ABC):
    @abstractmethod
    def get_dict(self) -> Dict:
        pass

    def load(self, config: 'BaseConfig'):
        for key, value in self.get_dict().items():
            self.get_setter().set(config, key, value)

    def get_sub_loader(self, val) -> "BaseConfigLoader":
        return DictConfigLoader(val, self._attribute_setter)


class PathBasedConfigLoader(BaseConfigLoader, ABC):
    _path: Path

    def __init__(self, path: Union[Path, str], setter: Union[attribute_setters.AttributeSetter, str] = None):
        super().__init__(setter)
        self._path = path if isinstance(path, Path) else Path(path)

    @property
    def path(self) -> Path:
        return self._path

    @path.setter
    def path(self, path: Path) -> None:
        self._path = path


class YamlConfigLoader(DictBasedConfigLoader, PathBasedConfigLoader):

    def get_dict(self) -> Dict:
        if not self.path.is_file():
            raise FileNotFoundError(str(self.path))
        stream = self.path.open()
        data = yaml.safe_load(stream)

        if data is None:
            data = {}

        stream.close()
        return data


class DictConfigLoader(DictBasedConfigLoader):
    _dict: Dict

    def __init__(self, _dict: Dict, setter: Union[attribute_setters.AttributeSetter, str] = None):
        super().__init__(setter)
        self._dict = _dict

    def get_dict(self) -> Dict:
        return self._dict


class NamespaceConfigLoader(DictBasedConfigLoader):
    _args: Namespace

    def __init__(self, args: Namespace, setter: Union[attribute_setters.AttributeSetter, str] = None):
        super().__init__(setter)
        self._args = args

    def get_dict(self) -> Dict:
        return vars(self._args)



import types
from inspect import *


# noinspection PyShadowingBuiltins
def getmembers(object, predicate=None, *, sort=True):
    """Return all members of an object as (name, value) pairs sorted by name.
    Optionally, only return members that satisfy a given predicate."""
    if isclass(object):
        mro = (object,) + getmro(object)
    else:
        mro = ()
    results = []
    processed = set()
    names = dir(object) if sort else ordered_dir(object)
    # :dd any DynamicClassAttributes to the list of names if object is a class;
    # this may result in duplicate entries if, for example, a virtual
    # attribute with the same name as a DynamicClassAttribute exists
    try:
        for base in object.__bases__:
            for k, v in base.__dict__.items():
                if isinstance(v, types.DynamicClassAttribute):
                    names.append(k)
    except AttributeError:
        pass
    for key in names:
        # First try to get the value via getattr.  Some descriptors don't
        # like calling their __get__ (see bug #1785), so fall back to
        # looking in the __dict__.
        try:
            value = getattr(object, key)
            # handle the duplicate key
            if key in processed:
                raise AttributeError
        except AttributeError:
            for base in mro:
                if key in base.__dict__:
                    value = base.__dict__[key]
                    break
            else:
                # could be a (currently) missing slot member, or a buggy
                # __dir__; discard and move on
                continue
        if not predicate or predicate(value):
            results.append((key, value))
        processed.add(key)
    if sort:
        results.sort(key=lambda pair: pair[0])
    return results


def ordered_dir(obj):
    if not isclass(obj):
        obj = type(obj)

    result = {}

    for mro in reversed(getmro(obj)):
        result.update(mro.__dict__)

    return result.keys()



from typing import Collection, List, Optional, Mapping, Sequence, Any, Dict, TYPE_CHECKING, Callable

from .. import ConfigStructureVisitor, ConfigListStructure, ConfigStructure, PlaceHolder

if TYPE_CHECKING:
    from .. import BaseConfig


class ToCollectionVisitor(ConfigStructureVisitor):
    recursive: bool
    set_circular_to_none: bool
    result_stack: List[Optional[Collection]]
    visited: Dict["HashableRef", Any]
    _filter: Callable[[PlaceHolder], bool]

    # noinspection PyShadowingBuiltins
    def __init__(self, recursive=True, set_circular_to_none=False, filter=None):
        self.recursive = recursive
        self.set_circular_to_none = set_circular_to_none
        self.result_stack = [None]
        self.visited = {}
        self._filter = filter

    def visit_config(self, structure: "BaseConfig") -> None:
        result = {}
        placeholders = structure.get_all_placeholders().values()
        if self._filter is not None:
            placeholders = filter(self._filter, placeholders)
        for placeholder in placeholders:
            if not placeholder.is_assigned(structure) or placeholder.hidden:
                continue
            self._resolve_value(structure[placeholder.name])
            result[placeholder.name] = self.result_stack.pop()
        self.result_stack[-1] = result

    def visit_config_list(self, structure: ConfigListStructure) -> None:
        result = []
        for value in structure:
            self._resolve_value(value)
            result.append(self.result_stack.pop())
        self.result_stack[-1] = result

    def get_result(self) -> Collection:
        assert len(self.result_stack) == 1
        return self.result_stack[0]

    def _resolve_value(self, value) -> None:
        if self.recursive:
            ref = HashableRef(value)
            if ref in self.visited:
                self.result_stack.append(
                    None if self.set_circular_to_none else self.visited[ref]
                )
            elif isinstance(value, ConfigStructure):
                self.result_stack.append(None)
                self.visited[ref] = None
                value.accept(self)
                self.visited[ref] = self.result_stack[-1]
            elif not isinstance(value, str) and isinstance(value, Collection):
                self.result_stack.append(None)
                self.visited[ref] = None
                self._visit_collection(value)
                self.visited[ref] = self.result_stack[-1]
            else:
                self.result_stack.append(value)
        else:
            self.result_stack.append(value)

    def _visit_collection(self, collection: Collection) -> None:
        if isinstance(collection, Mapping):
            entries = collection.items()
            result = {}
        elif isinstance(collection, Sequence):
            entries = enumerate(collection)
            result = [None] * len(collection)
        else:
            self.result_stack[-1] = collection
            return
        for key, value in entries:
            self._resolve_value(value)
            result[key] = self.result_stack.pop()
        self.result_stack[-1] = result


class HashableRef:
    obj: Any

    def __init__(self, obj: Any):
        self.obj = obj

    def __eq__(self, other: Any):
        if not isinstance(other, HashableRef):
            return False
        return self.obj is other.obj

    def __hash__(self):
        return id(self.obj)



from pprint import pprint
from typing import List, Dict, Optional

from fancy import config as cfg


# TODO write tests for accurate error messages
class MyConfigEmb(cfg.BaseConfig):
    x: int = cfg.Option(type=int)
    y: int = cfg.Option(type=int)
    i: bool = cfg.Option(type=bool)
    parent: "MyConfig" = cfg.PlaceHolder()


class MyConfig(cfg.BaseConfig):
    _a: int = cfg.Option(name="a", required=True, description="hi", type=int)
    b: float = cfg.Option(required=False, type=float)
    c: bool = cfg.Option(required=True, type=bool)
    li: List[List[MyConfigEmb]] = cfg.Option(type=[[MyConfigEmb]])
    n: Optional[int] = cfg.Option(type=int, required=True, nullable=True)

    lazy_li: List[MyConfigEmb] = cfg.Lazy(lambda c: c.li[0])
    placeholder_dict: Dict[int, MyConfigEmb] = cfg.PlaceHolder()
    child: MyConfigEmb = cfg.Option(type=MyConfigEmb)

    def post_load(self):
        self.placeholder_dict = {self._a: self.lazy_li[0]}

    @property
    def a(self):
        return self._a + 1


def test_config():
    opt = {
        "a": "3",
        "b": "4.2",
        "c": "true",
        'n': None,
        "li": [[{
            "x": 1,
            "y": 2,
            "i": "on",
            "a": "ig"
        },
            {
                "y": 3,
                "i": "off", }
        ]],
        "int_li": ["Y", "N"],
        "child": {
            "x": 4,
            "y": 5,
            "i": "off",
            "a": "ig"
        }
    }

    c = MyConfig(cfg.DictConfigLoader(opt, setter="ignore"))
    c.child.parent = c
    print(c)
    print(c.placeholder_dict)
    pprint(c.to_dict(True, load_lazies=True), indent=1, depth=4)



from .boolean import boolean, make_boolean
from .config import config
from .config_list import config_list
from .flag_string import flag_string
from .flag_container import flag_container
from .auto_process_typ import auto_process_typ
from .identical import identical



from typing import Any, AbstractSet

TRUE_CHOICES = {"on", "1", "true", "yes", "==TRUE=="}
FALSE_CHOICES = {"off", "0", "false", "no", "==FALSE=="}


def make_boolean(true_choices: AbstractSet = None, false_choices: AbstractSet = None):
    def _inner(value):
        return boolean(value, true_choices, false_choices)
    return _inner


def boolean(value: Any, true_choices: AbstractSet = None, false_choices: AbstractSet = None) -> bool:
    if true_choices is None:
        true_choices = TRUE_CHOICES
    if false_choices is None:
        false_choices = FALSE_CHOICES
    if isinstance(value, str):
        _value = value.strip()
        if _value in true_choices:
            return True
        if _value in false_choices:
            return False
        raise ValueError(
            f"unexpected value: {value}, expected choices:"
            f" {str(true_choices.union(false_choices))}"
        )
    return bool(value)



import pytest

from fancy import config as cfg
from fancy.config.utils import DispatcherError


class MyConfig(cfg.BaseConfig):
    a: int = cfg.Option(type=int)
    b: float = cfg.Option(type=float)
    c: int = cfg.Option(default=2, type=int)


DATA_WITH_DEFAULT = {"a": 1, "b": 2}
RESULT_DATA_WITH_DEFAULT = {"a": 1, "b": 2.0, "c": 2}
DATA = {"a": 1, "b": 2, "c": 3}


def test_loader_construction():
    assert MyConfig(cfg.DictConfigLoader(DATA_WITH_DEFAULT)).to_dict() == RESULT_DATA_WITH_DEFAULT
    assert MyConfig(cfg.DictConfigLoader(DATA)).to_dict() == DATA


def test_dict_construction():
    assert MyConfig(DATA_WITH_DEFAULT).to_dict() == RESULT_DATA_WITH_DEFAULT
    assert MyConfig(DATA).to_dict() == DATA


def test_kwarg_construction():
    assert MyConfig(**DATA_WITH_DEFAULT).to_dict() == RESULT_DATA_WITH_DEFAULT
    assert MyConfig(**DATA).to_dict() == DATA


def test_dispatch():
    with pytest.raises(DispatcherError):
        MyConfig(loader=1)
    with pytest.raises(DispatcherError):
        # noinspection PyTypeChecker
        MyConfig(1)
    with pytest.raises(DispatcherError):
        MyConfig(1, a=1, b=2)



from typing import TYPE_CHECKING, Callable, Any

from . import PlaceHolder

if TYPE_CHECKING:
    from . import BaseConfig


class Lazy(PlaceHolder):
    readonly: bool = True

    def __init__(
            self,
            fn: Callable[['BaseConfig'], Any],
            name: str = None,
            description: str = None,
            hidden: bool = False
    ):
        super().__init__(name, description, hidden)
        self.fn = fn

    def __get__(self, instance: 'BaseConfig', owner):
        if instance is None:
            return self

        if not super().is_assigned(instance):
            vars(instance)[self.__name__] = self.fn(instance)
        return vars(instance)[self.__name__]

    def is_assigned(self, instance) -> bool:
        return True



from typing import Type, Callable, TYPE_CHECKING


if TYPE_CHECKING:
    from .. import BaseConfig


def config(config_type: Type["BaseConfig"]) -> Callable:
    from .. import BaseConfig
    if not issubclass(config_type, BaseConfig):
        raise TypeError(f"config type {config_type} must be instance of BaseConfig")

    def _inner(_: dict) -> config_type:
        return config_type(loader=None)
    return _inner



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from . import BaseConfigLoader


class ConfigContext(ABC):
    @abstractmethod
    def get_loader(self) -> "BaseConfigLoader":
        """
        :raises exc.ContextNotLoadedError when loader does not exist.
        """



from typing import Callable

from . import boolean, config, config_list
from ..typing import UnProcType


def auto_process_typ(typ: UnProcType) -> Callable:
    from .. import BaseConfig  # lazy import
    if isinstance(typ, list):  # https://github.com/susautw/fancy-config/issues/4
        return config_list(typ[0])
    if isinstance(typ, type):
        if issubclass(typ, bool):
            return boolean
        if issubclass(typ, BaseConfig):
            return config(typ)
    if isinstance(typ, Callable):
        return typ
    else:
        raise ValueError("_type must be Callable or subclass of BaseConfig")



from typing import Callable

from ..typing import UnProcType


def config_list(_type: UnProcType) -> Callable:
    from .. import ConfigListStructure

    def _inner(_):
        return ConfigListStructure(_type)
    return _inner



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fancy.config import BaseConfig


# https://github.com/susautw/fancy-config/issues/3
class PlaceHolder:
    _config_name: str
    _description: str
    readonly: bool = False

    def __init__(self, name: str = None, description: str = None, hidden: bool = False):
        self._config_name = name
        self._description = "" if description is None else description
        self.hidden = hidden

    def __set_name__(self, owner, name):
        self.__name__ = name
        if self._config_name is None:
            self._config_name = name

    def __set__(self, instance: 'BaseConfig', raw_value):
        if self.readonly:
            raise AttributeError(f"{self.name} can't be set")
        vars(instance)[self.__name__] = raw_value

    @property
    def name(self) -> str:
        return self._config_name

    @property
    def description(self) -> str:
        return self._description

    def is_assigned(self, instance) -> bool:
        """
        :return true, if the placeholder is accessible.
        """
        return self.__name__ in vars(instance)



from typing import Any, Callable

from . import ConfigStructure, ConfigContext, ConfigStructureVisitor
from .process import auto_process_typ
from .typing import UnProcType


class ConfigListStructure(list, ConfigStructure):
    _config_typ: Callable[[Any], Any]

    def __init__(self, config_typ: UnProcType):
        super().__init__()
        self._config_typ = auto_process_typ(config_typ)

    def load_by_context(self, context: ConfigContext, val):
        new_items = []
        for raw_cfg in val:
            cfg = self._config_typ(raw_cfg)
            if isinstance(cfg, ConfigStructure):
                cfg.load_by_context(context, raw_cfg)
            new_items.append(cfg)
        self.extend(new_items)

    def accept(self, visitor: "ConfigStructureVisitor"):
        visitor.visit_config_list(self)



from . import attribute_setters

from . import process
from .process import (
    config_list,
    make_boolean,
    identical
)

from .config_context import ConfigContext
from .config_structure import ConfigStructure
from .config_structure_visitor import ConfigStructureVisitor
from .placeholder import PlaceHolder
from .lazy import Lazy
from .option import Option
from .config_list_struct import ConfigListStructure
from .config_loaders import (
    BaseConfigLoader,
    DictBasedConfigLoader,
    PathBasedConfigLoader,
    YamlConfigLoader,
    DictConfigLoader
)
from .base_config import BaseConfig


from .config_loader_factory import ConfigLoaderFactory
from .config_factory import ConfigFactory



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from fancy.config import BaseConfig


class AttributeSetter(ABC):
    @abstractmethod
    def set(self, config: "BaseConfig", key: str, val: Any) -> None:
        pass

    @classmethod
    def get_setter_name(cls):
        raise NotImplementedError()


class StrictAttributeSetter(AttributeSetter):
    def set(self, config: "BaseConfig", key: str, val: Any) -> None:
        """
        :raise ValueError, TypeError
        """
        config[key] = val

    @classmethod
    def get_setter_name(cls):
        return "strict"


class IgnoreErrorAttributeSetter(AttributeSetter):
    """
    :raise ValueError, TypeError
    """

    def set(self, config: "BaseConfig", key: str, val: Any) -> None:
        try:
            config[key] = val
        except KeyError:
            pass

    @classmethod
    def get_setter_name(cls):
        return "ignore"



from abc import ABC
import warnings
from typing import Dict, List, overload, Any, Callable

from . import ConfigStructure, ConfigContext, exc, PlaceHolder, ConfigStructureVisitor, DictConfigLoader
from . import Option
from .utils import inspect, Dispatcher, DispatcherError
from . import visitors
from ..config import BaseConfigLoader


class BaseConfig(ConfigStructure, ConfigContext, ABC):
    _name_mapping: Dict[str, str] = None
    _all_placeholders: Dict[str, PlaceHolder] = None
    _all_options: Dict[str, Option] = None
    _all_required_options: List[Option] = None
    _loader: BaseConfigLoader = None

    _method_init_ = Dispatcher(is_method=True)

    @overload
    @_method_init_.register
    def __init__(self, loader: BaseConfigLoader = None):
        if loader is not None:
            if not isinstance(loader, BaseConfigLoader):
                raise DispatcherError("loader should be a BaseConfigLoader")
            self.load(loader)

    @overload
    @_method_init_.register
    def __init__(self, options: Dict[str, Any]):
        if not isinstance(options, dict):
            raise DispatcherError("options should be dict")
        self.__init__(DictConfigLoader(options))

    @overload
    @_method_init_.register
    def __init__(self, **options: Any):
        if "loader" in options and len(options) == 1:
            raise DispatcherError("loader should be a BaseConfigLoader")
        self.__init__(options)

    @_method_init_.implement
    def __init__(self, *args, **kwargs):
        ...

    def load(self, loader: 'BaseConfigLoader') -> None:
        self._loader = loader
        loader.load(self)
        self._postprocessing()
        self.post_load()

    def _postprocessing(self) -> None:
        for option in self.get_all_required_options():
            if not hasattr(self, option.__name__):
                raise ValueError(f'{type(self)}: the missing placeholder {option.name} is required.')

    def load_by_context(self, context: ConfigContext, val):
        self.load(context.get_loader().get_sub_loader(val))

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise TypeError(f'{type(self)}: {item} must be str, not {type(item)}')
        try:
            return self.__getattribute__(self.get_name_mapping()[item])
        except AttributeError:
            raise KeyError(f'{type(self)}: not contains the config named {item}')

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError(f'{type(self)}: {key} must be str, not {type(key)}')
        if key not in self.get_name_mapping():
            raise KeyError(f'{type(self)}: not contains the config named {key}, value: {repr(value)}')
        key = self.get_name_mapping()[key]
        self.__setattr__(key, value)

    def get_loader(self) -> 'BaseConfigLoader':
        if self._loader is None:
            raise exc.ContextNotLoadedError(self)
        return self._loader

    @property
    def loaded(self) -> bool:
        return self._loader is not None

    def post_load(self):
        pass

    # noinspection PyShadowingBuiltins
    def to_dict(
            self,
            recursive=True,
            prevent_circular=False, *,
            load_lazies=None,
            filter: Callable[[PlaceHolder], bool] = None
    ) -> dict:
        """
        convert this config to a dictionary
        :param recursive: If true, the method will convert structures in this config recursively.
        :param prevent_circular: If true, the method will set the circular instance to `None` in the result.
        :param filter: a Callable to know what placeholders should be used.
        :param load_lazies: Deprecated since 0.12.0
        :return:
        """
        if load_lazies is not None:
            warnings.warn(
                "the parameter 'load_lazies' is deprecated and will be removed in 1.0.0.",
                DeprecationWarning
            )
        visitor = visitors.ToCollectionVisitor(
            recursive=recursive, set_circular_to_none=prevent_circular, filter=filter
        )
        self.accept(visitor)

        # noinspection PyTypeChecker
        #  The visitor must return a dictionary
        return visitor.get_result()

    def accept(self, visitor: "ConfigStructureVisitor"):
        visitor.visit_config(self)

    def load_lazies(self) -> None:
        warnings.warn(
            "This method is deprecated and will be removed in 1.0.0."
            "And the method has no operation",
            DeprecationWarning
        )

    def __repr__(self):
        return str(self.to_dict())

    def __str__(self):
        return self.__repr__()

    @classmethod
    def get_all_placeholders(cls) -> Dict[str, PlaceHolder]:
        if cls._all_placeholders is None:
            cls._all_placeholders = {
                name: placeholder
                for name, placeholder in inspect.getmembers(
                    cls, lambda placeholder: isinstance(placeholder, PlaceHolder), sort=False
                )
            }
        return cls._all_placeholders

    @classmethod
    def get_all_options(cls) -> Dict[str, Option]:
        if cls._all_options is None:
            cls._all_options = {
                name: option
                for name, option in cls.get_all_placeholders().items() if isinstance(option, Option)
            }
        return cls._all_options

    @classmethod
    def get_all_required_options(cls) -> List[Option]:
        if cls._all_required_options is None:
            cls._all_required_options = [option for option in cls.get_all_options().values() if option.required]
        return cls._all_required_options

    @classmethod
    def get_name_mapping(cls) -> Dict[str, str]:
        if cls._name_mapping is None:
            cls._name_mapping = {
                placeholder.name: attr_name
                for attr_name, placeholder in cls.get_all_placeholders().items()
            }
        return cls._name_mapping



import functools
import inspect
from typing import Dict, Callable


class Dispatcher:
    _registry: Dict[Callable, inspect.Signature] = {}
    _is_method: bool

    def __init__(self, is_method=False):
        self._is_method = is_method

    def register(self, method):
        self._registry[method] = inspect.signature(method)
        return method

    def dispatch(self, args, kwargs):
        errs = []
        for m, sig in self._registry.items():
            try:
                ba = self._bind_sig(sig, args, kwargs)
                ba.apply_defaults()
                return m(*ba.args, **ba.kwargs)
            except DispatcherError as e:
                errs.append(e)
        raise DispatcherError("No signature matched", errs)

    def implement(self, handler):
        @functools.wraps(handler)
        def _wrapper(*args, **kwargs):
            try:
                self.dispatch(args, kwargs)
                self._exec_handler(handler, None, args, kwargs)
            except DispatcherError as e:
                if not self._exec_handler(handler, e, args, kwargs):
                    raise

        return _wrapper

    def _bind_sig(self, sig, args, kwargs):
        try:
            ba = sig.bind(*args, **kwargs)
            return ba
        except TypeError as e:
            raise DispatcherError(e.args[0])

    def _exec_handler(self, handler, e, args, kwargs):
        if self._is_method:
            _self, *args = args
            return handler(_self, e, *args, **kwargs)
        else:
            return handler(e, *args, **kwargs)


class DispatcherError(TypeError):
    pass



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from . import ConfigContext

if TYPE_CHECKING:
    from . import ConfigStructureVisitor


class ConfigStructure(ABC):
    @abstractmethod
    def load_by_context(self, context: ConfigContext, val): ...

    @abstractmethod
    def accept(self, visitor: "ConfigStructureVisitor"): ...



from typing import Optional, List

from fancy import config as cfg


def test_config_to_dict_with_default_values():
    class MySubConfig(cfg.BaseConfig):
        x: bool = cfg.Option(default=True, type=bool)
        y: Optional[float] = cfg.Option(nullable=True, type=float)

    class MyConfig(cfg.BaseConfig):
        a: int = cfg.Option(type=int)
        b: float = cfg.Option(default=5.1, type=float)
        sub: MySubConfig = cfg.Option(default={}, type=MySubConfig)

    assert MyConfig(a=1).to_dict() == {
        'a': 1,
        'b': 5.1,
        'sub': {'x': True, 'y': None}
    }


def test_to_dict_with_lazy_placeholder():
    class SubConfig(cfg.BaseConfig):
        x: int = cfg.Option(type=int)
        y: bool = cfg.Lazy(lambda c: c.x == 0)

    class MyConfig(cfg.BaseConfig):
        a: int = cfg.Option(type=int)
        b: float = cfg.Lazy(lambda c: c.a / 2)
        sub: SubConfig = cfg.Option(default={}, type=SubConfig)
        sub_list: List[SubConfig] = cfg.Option(default=[], type=[SubConfig])

    assert MyConfig(
        a=1, sub=dict(x=1), sub_list=[dict(x=0), dict(x=1)]
    ).to_dict() == {
        'a': 1, 'b': 0.5,
        'sub': {'x': 1, 'y': False},
        'sub_list': [{'x': 0, 'y': True}, {'x': 1, 'y': False}]
    }


def test_to_dict_with_filter():
    class MyConfig(cfg.BaseConfig):
        a: int = cfg.Option(default=0, type=int)
        b: float = cfg.PlaceHolder()

        def post_load(self):
            self.b = self.a * 1.2

    c = MyConfig(a=1)
    assert c.to_dict(filter=lambda p: isinstance(p, cfg.Option)) == {"a": 1}
    assert c.to_dict(filter=lambda p: p.name != "a") == {"b": 1.2}


def test_to_dict_with_hidden_placeholder():
    class MyConfig(cfg.BaseConfig):
        a: int = cfg.Option(type=int)
        _b: int = cfg.Option(name="b", type=int, hidden=True)
        _c: int = cfg.Lazy(lambda _c: 40, hidden=True)
        _d: int = cfg.PlaceHolder(hidden=True)

        def post_load(self):
            self._d = 20

    c = MyConfig(a=1, b=2)
    assert c.to_dict() == {"a": 1}
    assert c.a == 1
    assert c._b == 2
    assert c._c == 40
    assert c._d == 20



from typing import Any, Callable, TYPE_CHECKING

from . import ConfigStructure, PlaceHolder
from .process import auto_process_typ
from .typing import UnProcType
from ..config import identical

if TYPE_CHECKING:
    from ..config import BaseConfig


class Option(PlaceHolder):
    raw_type: UnProcType
    _type: Callable[[Any], Any]
    _required: bool
    _nullable: bool
    _default: Any

    _description: str

    _config_name: str = None
    readonly: bool = False

    def __init__(
            self,
            required: bool = False,
            nullable: bool = False,
            default=None,
            type: UnProcType = identical,
            name: str = None,
            description: str = None,
            hidden: bool = False
    ):
        super().__init__(name, description, hidden)
        self._required = required
        self._nullable = nullable

        # TODO resolve default value problem
        #  if type is a Config or ConfigList and "Option isn't required or nullable", the
        #  default value should assigned a empty dict or list in order to transfer control to sub configs
        #  ---
        #  the "transfer control" means that sub config decides what the default value
        #  or required value or nullable values
        self._default = default

        self.raw_type = type
        self._type = auto_process_typ(type)

    def __get__(self, instance: 'BaseConfig', owner):
        if instance is None:
            return self

        # initialize value
        if self._should_assign_default_value(instance):
            if self._default is None and not self._nullable:
                raise AttributeError("attribute must assign the value before access it.")

            self.__set__(instance, self._default)

        return vars(instance)[self.__name__]

    def __set__(self, instance, raw_value):
        # TODO Add Docs to explain how the None value works in this function
        if raw_value is None:
            if not self._nullable:
                raise ValueError('the value should not be none')
            value = None
        else:
            value = self._type(raw_value)
        if isinstance(value, ConfigStructure):
            value.load_by_context(instance, raw_value)

        vars(instance)[self.__name__] = value

    def __delete__(self, instance):
        del vars(instance)[self.__name__]

    def _should_assign_default_value(self, instance):
        return not super().is_assigned(instance)

    @property
    def required(self) -> bool:
        return self._required

    def is_assigned(self, instance) -> bool:
        return super().is_assigned(instance) or self._default is not None or self._nullable



from copy import deepcopy
from typing import TypeVar, MutableMapping, MutableSequence

from . import flag_string

T = TypeVar("T", MutableMapping, MutableSequence)


def flag_container(container: T) -> T:
    if not isinstance(container, (MutableMapping, MutableSequence)):
        raise TypeError("the container must be a MutableMapping or MutableSequence")
    result = deepcopy(container)
    opening = [result]

    while len(opening) > 0:
        container = opening.pop()
        if isinstance(container, MutableMapping):
            keys = container.keys()
        else:
            keys = range(len(container))

        for k in keys:
            if isinstance(container[k], (MutableMapping, MutableSequence)):
                opening.append(container[k])
            else:
                container[k] = flag_string(container[k])
    return result
